from .parameter import parameter
from .pipeline import pipeline
from .run import current_run
from .runtime import download_pipeline, import_pipeline
from .utils import get_local_workspace_config

__all__ = [
    "pipeline",
    "parameter",
    "current_run",
    "import_pipeline",
    "download_pipeline",
    "get_local_workspace_config",
]



import enum
import os


class Environments(enum.Enum):
    LOCAL = "LOCAL"
    PIPELINE = "PIPELINE"
    DOCKER = "DOCKER"
    JUPYTER = "JUPYTER"


def get_environment():
    env = os.environ.get("HEXA_ENVIRONMENT", "LOCAL").upper()
    if env not in Environments.__members__:
        raise ValueError(f"Invalid environment: {env}")
    return Environments[env]



import os
from unittest.mock import Mock, patch

from openhexa.sdk import current_run, workspace
from openhexa.sdk.pipelines.pipeline import Pipeline


@patch.dict(os.environ, {"WORKSPACE_FILES_PATH": "/workspace/path"})
def test_workspace_local_files_path():
    assert workspace.files_path == "/workspace/path"


@patch.dict(os.environ, {"HEXA_ENVIRONMENT": "PIPELINE"})
def test_workspace_pipeline_files_path():
    assert workspace.files_path == "/home/hexa/workspace"


@patch.dict(
    os.environ,
    {"HEXA_ENVIRONMENT": "PIPELINE", "HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_connected():
    assert current_run.connected is True


@patch.dict(
    os.environ,
    {"HEXA_ENVIRONMENT": "LOCAL", "HEXA_SERVER_URL": "https://test.openhexa.org"},
)
def test_not_connected():
    assert current_run.connected is False


@patch.dict(
    os.environ,
    {
        "HEXA_ENVIRONMENT": "LOCAL",
    },
)
def test_not_connected_missing_url():
    assert current_run.connected is False


@patch.dict(
    os.environ,
    {
        "HEXA_ENVIRONMENT": "LOCAL",
    },
)
def test_pipeline_local():
    pipeline_func = Mock()

    pipeline = Pipeline("code", "pipeline", pipeline_func, [])
    assert pipeline.connected is False


@patch.dict(
    os.environ,
    {
        "HEXA_ENVIRONMENT": "PIPELINE",
        "HEXA_SERVER_URL": "https://test.openhexa.org",
    },
)
def test_pipeline_pipeline():
    pipeline_func = Mock()

    pipeline = Pipeline("code", "pipeline", pipeline_func, [])
    assert pipeline.connected is True



import re
import typing


class ParameterValueError(Exception):
    pass


class ParameterType:
    """Base class for parameter types. Those parameter types are used when using the @parameter decorator"""

    def spec_type(self) -> str:
        """Returns a type string for the specs that are sent to the backend."""

        raise NotImplementedError

    @property
    def expected_type(self) -> typing.Type:
        """Returns the python type expected for values."""

        raise NotImplementedError

    @property
    def accepts_choice(self) -> bool:
        return True

    @property
    def accepts_multiple(self) -> bool:
        return True

    @staticmethod
    def normalize(value: typing.Any) -> typing.Any:
        """If appropriate, subclasses can override this method to normalize empty values to None.

        This can be used to handle empty values and normalize them to None, or to perform type conversions, allowing us
        to allow multiple input types but still normalize everything to a single type.
        """

        return value

    def validate(
        self, value: typing.Optional[typing.Any], allow_empty: bool = True
    ) -> typing.Optional[typing.Any]:
        """Validate the provided value for this type."""

        if not isinstance(value, self.expected_type):
            raise ParameterValueError(
                f"Invalid type for value {value} (expected {self.expected_type}, got {type(value)})"
            )

        return value

    def __str__(self) -> str:
        return str(self.expected_type)


class String(ParameterType):
    @property
    def spec_type(self) -> str:
        return "str"

    @property
    def expected_type(self) -> typing.Type:
        return str

    @staticmethod
    def normalize(value: typing.Any) -> typing.Optional[str]:
        if isinstance(value, str):
            normalized_value = value.strip()
        else:
            normalized_value = value

        if normalized_value == "":
            return None

        return normalized_value

    def validate(
        self, value: typing.Optional[typing.Any], *, allow_empty: bool = True
    ) -> typing.Optional[str]:
        if not allow_empty and value == "":
            raise ParameterValueError("Empty values are not accepted.")

        return super().validate(value, allow_empty)


class Boolean(ParameterType):
    @property
    def spec_type(self) -> str:
        return "bool"

    @property
    def expected_type(self) -> typing.Type:
        return bool

    @property
    def accepts_choice(self) -> bool:
        return False

    @property
    def accepts_multiple(self) -> bool:
        return False


class Integer(ParameterType):
    @property
    def spec_type(self) -> str:
        return "int"

    @property
    def expected_type(self) -> typing.Type:
        return int


class Float(ParameterType):
    @property
    def spec_type(self) -> str:
        return "float"

    @property
    def expected_type(self) -> typing.Type:
        return float

    @staticmethod
    def normalize(value: typing.Any) -> typing.Any:
        if isinstance(value, int):
            return float(value)

        return value


TYPES_BY_PYTHON_TYPE = {str: String, bool: Boolean, int: Integer, float: Float}


class InvalidParameterError(Exception):
    pass


class Parameter:
    """Pipeline parameter class. Contains validation logic specs generation logic."""

    def __init__(
        self,
        code: str,
        *,
        type: typing.Union[typing.Type[str], typing.Type[int], typing.Type[bool]],
        name: typing.Optional[str] = None,
        choices: typing.Optional[typing.Sequence] = None,
        help: typing.Optional[str] = None,
        default: typing.Optional[typing.Any] = None,
        required: bool = True,
        multiple: bool = False,
    ):
        if re.match("^[a-z_][a-z_0-9]+$", code) is None:
            raise InvalidParameterError(
                f"Invalid parameter code provided ({code}). Parameter must start with a letter or an underscore, and can only contain lower case letters, numbers and underscores."
            )

        self.code = code

        try:
            self.type = TYPES_BY_PYTHON_TYPE[type]()
        except KeyError:
            valid_parameter_types = [str(k) for k in TYPES_BY_PYTHON_TYPE.keys()]
            raise InvalidParameterError(
                f"Invalid parameter type provided ({type}). Valid parameter types are {', '.join(valid_parameter_types)}"
            )

        if choices is not None:
            if not self.type.accepts_choice:
                raise InvalidParameterError(
                    f"Parameters of type {self.type} don't accept choices."
                )
            elif len(choices) == 0:
                raise InvalidParameterError("Choices, if provided, cannot be empty.")

            try:
                for choice in choices:
                    self.type.validate(choice)
            except ParameterValueError:
                raise InvalidParameterError(
                    f"The provided choices are not valid for the {self.type} parameter type."
                )
        self.choices = choices

        self.name = name
        self.help = help
        self.required = required

        if multiple is True and not self.type.accepts_multiple:
            raise InvalidParameterError(
                f"Parameters of type {self.type} can't have multiple values."
            )
        self.multiple = multiple

        self._validate_default(default, multiple)
        self.default = default

    def validate(self, value: typing.Any) -> typing.Any:
        """Validates the provided value against the parameter, taking required / default options into account."""

        if self.multiple:
            return self._validate_multiple(value)
        else:
            return self._validate_single(value)

    def _validate_single(self, value: typing.Any):
        # Normalize empty values to None and handles default
        normalized_value = self.type.normalize(value)
        if normalized_value is None and self.default is not None:
            normalized_value = self.default

        if normalized_value is None:
            if self.required:
                raise ParameterValueError(f"{self.code} is required")

            return None

        pre_validated = self.type.validate(normalized_value)
        if self.choices is not None and pre_validated not in self.choices:
            raise ParameterValueError(
                f"The provided value for {self.code} is not included in the provided choices."
            )

        return pre_validated

    def _validate_multiple(self, value: typing.Any):
        # Reject values that are not lists
        if value is not None and not isinstance(value, list):
            raise InvalidParameterError(
                "If provided, value should be a list when parameter is multiple."
            )

        # Normalize empty values to an empty list
        if value is None:
            normalized_value = []
        else:
            normalized_value = [self.type.normalize(v) for v in value]
            normalized_value = list(filter(lambda v: v is not None, normalized_value))
        if len(normalized_value) == 0 and self.default is not None:
            normalized_value = self.default

        if len(normalized_value) == 0 and self.required:
            raise ParameterValueError(f"{self.code} is required")

        pre_validated = [
            self.type.validate(single_value) for single_value in normalized_value
        ]
        if self.choices is not None and any(
            v not in self.choices for v in pre_validated
        ):
            raise ParameterValueError(
                f"One of the provided values for {self.code} is not included in the provided choices."
            )

        return pre_validated

    def _validate_default(self, default: typing.Any, multiple: bool):
        if default is None:
            return

        try:
            if multiple:
                if not isinstance(default, list):
                    raise InvalidParameterError(
                        "Default values should be lists when using multiple=True"
                    )
                for default_value in default:
                    self.type.validate(default_value, allow_empty=False)
            else:
                self.type.validate(default, allow_empty=False)
        except ParameterValueError:
            raise InvalidParameterError(
                f"The default value for {self.code} is not valid."
            )

    def parameter_spec(self):
        """Generates specification for this parameter, to be provided to the OpenHexa backend."""

        return {
            "type": self.type.spec_type,
            "required": self.required,
            "choices": self.choices,
            "code": self.code,
            "name": self.name,
            "help": self.help,
            "multiple": self.multiple,
            "default": self.default,
        }


def parameter(
    code: str,
    *,
    type: typing.Union[
        typing.Type[str], typing.Type[int], typing.Type[bool], typing.Type[float]
    ],
    name: typing.Optional[str] = None,
    choices: typing.Optional[typing.Sequence] = None,
    help: typing.Optional[str] = None,
    default: typing.Optional[typing.Any] = None,
    required: bool = True,
    multiple: bool = False,
):
    """parameter decorator."""

    def decorator(fun):
        return FunctionWithParameter(
            fun,
            Parameter(
                code,
                type=type,
                name=name,
                choices=choices,
                help=help,
                default=default,
                required=required,
                multiple=multiple,
            ),
        )

    return decorator


class FunctionWithParameter:
    """This class serves as a wrapper for functions decorated with the @parameter decorator."""

    def __init__(self, function, added_parameter: Parameter):
        self.function = function
        self.parameter = added_parameter

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def get_all_parameters(self):
        if isinstance(self.function, FunctionWithParameter):
            return [self.parameter, *self.function.get_all_parameters()]

        return [self.parameter]



from pathlib import Path

import stringcase
import yaml


class LocalWorkspaceConfigError(Exception):
    pass


def get_local_workspace_config(path: Path):
    """Get local workspace config for development/testing purposes.

    As of now, this function attempts to make the local run context looks like an online run context by
    filling environment variables.

    This is obviously brittle as it relies on setting the correct env variables keys, any changes upstream must
    be reflected here.
    """

    env_vars = {}
    # This will only work when running the pipeline using "python pipeline.py"
    # (We will have to find another approach for tests or running the pipeline using the CLI)
    local_workspace_config_path = path / Path("workspace.yaml")
    if not local_workspace_config_path.exists():
        print("Does not exist")
        return

    with open(
        local_workspace_config_path.resolve(), "r"
    ) as local_workspace_config_file:
        local_workspace_config = yaml.safe_load(local_workspace_config_file)
        # Database config
        if "database" in local_workspace_config:
            try:
                env_vars["WORKSPACE_DATABASE_USERNAME"] = local_workspace_config[
                    "database"
                ]["username"]
                password = local_workspace_config["database"].get("password")
                if password is not None:
                    env_vars["WORKSPACE_DATABASE_PASSWORD"] = password
                env_vars["WORKSPACE_DATABASE_HOST"] = local_workspace_config[
                    "database"
                ]["host"]
                env_vars["WORKSPACE_DATABASE_PORT"] = str(
                    local_workspace_config["database"]["port"]
                )
                env_vars["WORKSPACE_DATABASE_DB_NAME"] = local_workspace_config[
                    "database"
                ]["dbname"]
            except KeyError:
                exception_message = (
                    "Invalid local workspace database config. Please make sure you provide the following "
                    "keys: database.username, database.password, database.host, database.port, "
                    "database.dbname."
                )
                raise LocalWorkspaceConfigError(exception_message)

        # Files config
        if "files" in local_workspace_config:
            try:
                files_path = path / Path(local_workspace_config["files"]["path"])
                if not files_path.exists():
                    raise LocalWorkspaceConfigError(
                        f"The {files_path} files path does not exist. "
                        f"Did you forget to create it?"
                    )
                env_vars["WORKSPACE_FILES_PATH"] = str(files_path.resolve())
            except KeyError:
                exception_message = (
                    "Invalid local workspace database config. Please make sure you provide the following "
                    "keys: username, password, host, port, name."
                )
                raise LocalWorkspaceConfigError(exception_message)

        # Connections
        if "connections" in local_workspace_config:
            for slug, connection_config in local_workspace_config[
                "connections"
            ].items():
                if "type" not in connection_config:
                    raise LocalWorkspaceConfigError(
                        "Each connection must have a type key."
                    )

                # DHIS2 connections
                if connection_config["type"] == "dhis2":
                    try:
                        env_vars[
                            f"{stringcase.constcase(slug)}_URL"
                        ] = connection_config["url"]
                        env_vars[
                            f"{stringcase.constcase(slug)}_USERNAME"
                        ] = connection_config["username"]
                        env_vars[
                            f"{stringcase.constcase(slug)}_PASSWORD"
                        ] = connection_config["password"]
                    except KeyError:
                        exception_message = (
                            "Invalid local workspace dhis2 connection config. Please make sure you provide "
                            "the following keys: url, username, password."
                        )
                        raise LocalWorkspaceConfigError(exception_message)
                # PostgreSQL connections
                elif connection_config["type"] == "postgresql":
                    try:
                        env_vars[
                            f"{stringcase.constcase(slug)}_HOST"
                        ] = connection_config["host"]
                        env_vars[f"{stringcase.constcase(slug)}_PORT"] = str(
                            connection_config["port"]
                        )
                        env_vars[f"{stringcase.constcase(slug)}_USERNAME"] = str(
                            connection_config["username"]
                        )
                        env_vars[
                            f"{stringcase.constcase(slug)}_PASSWORD"
                        ] = connection_config["password"]
                        env_vars[
                            f"{stringcase.constcase(slug)}_DB_NAME"
                        ] = connection_config["database_name"]
                    except KeyError:
                        exception_message = (
                            "Invalid local workspace PostgreSQL connection config. Please make sure you provide "
                            "the following keys: url, username, password."
                        )
                        raise LocalWorkspaceConfigError(exception_message)
                # S3 connections
                elif connection_config["type"] == "s3":
                    try:
                        env_vars[
                            f"{stringcase.constcase(slug)}_SECRET_ACCESS_KEY"
                        ] = connection_config["secret_access_key"]
                        env_vars[f"{stringcase.constcase(slug)}_ACCESS_KEY_ID"] = str(
                            connection_config["access_key_id"]
                        )
                        env_vars[f"{stringcase.constcase(slug)}_BUCKET_NAME"] = str(
                            connection_config["bucket_name"]
                        )
                    except KeyError:
                        exception_message = (
                            "Invalid local workspace S3 connection config. Please make sure you provide "
                            "the following keys: secret_key, access_key_id, bucket_name."
                        )
                        raise LocalWorkspaceConfigError(exception_message)
                # GCS connections
                elif connection_config["type"] == "gcs":
                    try:
                        env_vars[
                            f"{stringcase.constcase(slug)}_SERVICE_ACCOUNT_KEY"
                        ] = connection_config["service_account_key"]
                        env_vars[f"{stringcase.constcase(slug)}_BUCKET_NAME"] = str(
                            connection_config["bucket_name"]
                        )
                    except KeyError:
                        exception_message = (
                            "Invalid local workspace GCS connection config. Please make sure you provide "
                            "the following keys: service_account_key, bucket_name."
                        )
                        raise LocalWorkspaceConfigError(exception_message)
                # Custom connection
                else:
                    for key, value in connection_config.items():
                        if key != "type":
                            env_vars[stringcase.constcase(f"{slug}_{key}")] = str(value)
    return env_vars



from __future__ import annotations

import argparse
import datetime
import json
import os
import string
import sys
import time
import typing
from logging import getLogger
from pathlib import Path

import requests
from multiprocess import get_context  # NOQA

from openhexa.sdk.utils import Environments, get_environment

from .parameter import FunctionWithParameter, Parameter, ParameterValueError
from .task import PipelineWithTask
from .utils import get_local_workspace_config

logger = getLogger(__name__)


class PipelineConfigError(Exception):
    pass


def pipeline(
    code: str, *, name: str = None
) -> typing.Callable[[typing.Callable[..., typing.Any]], "Pipeline"]:
    if any(c not in string.ascii_lowercase + string.digits + "_-" for c in code):
        raise Exception(
            "Pipeline name should contains only lower case letters, digits, '_' and '-'"
        )

    def decorator(fun):
        if isinstance(fun, FunctionWithParameter):
            parameters = fun.get_all_parameters()
        else:
            parameters = []

        return Pipeline(code, name, fun, parameters)

    return decorator


class PipelineRunError(Exception):
    pass


class Pipeline:
    def __init__(
        self,
        code: str,
        name: str,
        function: typing.Callable,
        parameters: typing.Sequence[Parameter],
    ):
        self.code = code
        self.name = name
        self.function = function
        self.parameters = parameters
        self.tasks = []

    def task(self, function):
        """task decorator"""

        return PipelineWithTask(function, self)

    def run(self, config: typing.Dict[str, typing.Any]):
        now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()
        print(f'{now} Starting pipeline "{self.code}"')

        # Validate / default parameters
        validated_config = {}
        for parameter in self.parameters:
            value = config.pop(parameter.code, None)
            validated_value = parameter.validate(value)
            validated_config[parameter.code] = validated_value

        if len(config) > 0:
            raise ParameterValueError(
                f"The provided config contains invalid key(s): {', '.join(list(config.keys()))}"
            )

        self.function(**validated_config)

        # managing variables
        result_list = []
        context = get_context("spawn")
        pool = context.Pool()  # FIXME: set max size of pool
        total = len(self.tasks)
        completed = 0

        while True:
            tasks = self.get_available_tasks()
            # filter already pooled task, even if they are not finished
            tasks = [t for t in tasks if not t.pooled]

            if len(tasks) == 0 and len(result_list) == 0:
                # nothing running, nothing to do
                break

            for task in tasks:
                now = datetime.datetime.utcnow().replace(microsecond=0).isoformat()
                print(f'{now} Started task "{task.compute.__name__}"')
                result = pool.apply_async(task.run)
                result_list.append((result, task))
                task.pooled = True

            while True:
                dag_step = False

                for result, task in result_list:
                    if not result.ready():
                        continue
                    now = (
                        datetime.datetime.now(tz=datetime.timezone.utc)
                        .replace(microsecond=0)
                        .isoformat()
                    )

                    completed += 1
                    progress = int(completed / total * 100)
                    print(f'{now} Finished task "{task.compute.__name__}"')
                    self._update_progress(progress)

                    try:
                        task_com_result = result.get()
                        task.result = task_com_result.result
                        task.start_time = task_com_result.start_time
                        task.end_time = task_com_result.end_time
                        dag_step = True
                    except Exception as e:  # NOQA
                        raise PipelineRunError(f"Pipeline {self.code} failed: {e}")

                if dag_step:
                    # remove finished tasks
                    result_list = [(r, t) for r, t in result_list if t.end_time is None]
                    break
                else:
                    # busy loop
                    time.sleep(0.3)

        pool.close()
        pool.join()

        print(f'{now} Successfully completed pipeline "{self.code}"')

    def get_available_tasks(self):
        return [task for task in self.tasks if task.is_ready()]

    def parameters_spec(self):
        return [arg.parameter_spec() for arg in self.parameters]

    def _update_progress(self, progress: int):
        if self.connected:
            token = os.environ["HEXA_TOKEN"]
            headers = {"Authorization": "Bearer %s" % token}
            query = """
                            mutation updatePipelineProgress ($input: UpdatePipelineProgressInput!) {
                                updatePipelineProgress(input: $input) { success errors }
                            }"""
            r = requests.post(
                f'{os.environ["HEXA_SERVER_URL"]}/graphql/',
                headers=headers,
                json={
                    "query": query,
                    "variables": {"input": {"percent": progress}},
                },
            )
            r.raise_for_status()
        else:
            print(f"Progress update: {progress}%")

    @property
    def connected(self):
        env = get_environment()
        return env == Environments.PIPELINE and "HEXA_SERVER_URL" in os.environ

    def __call__(self, config: typing.Optional[typing.Dict[str, typing.Any]] = None):
        # Handle local workspace config for dev / testing, if appropriate
        if get_environment() == Environments.LOCAL:
            os.environ.update(get_local_workspace_config(Path(sys.argv[0]).parent))

        # Handle config
        if config is None:  # Called without arguments, in the pipeline file itself
            parser = argparse.ArgumentParser(exit_on_error=False)
            parser.add_argument("-c", "--config")
            parser.add_argument("-f", "--config-file")
            # We can't use parse_args, as it will call sys.exit() if there are unrecognized arguments
            args, argv = parser.parse_known_args()
            if argv or (args.config_file is not None and args.config is not None):
                raise ValueError(
                    f"Unrecognized arguments: {' '.join(argv)}. Running a pipeline requires a single "
                    "argument: either an inline JSON config with the --values/-v argument, or a JSON "
                    "config file with the --values-file/-f argument."
                )
            if args.config_file is not None:
                with open(args.config_file, "r") as cf:
                    try:
                        config = json.load(cf)
                    except json.JSONDecodeError:
                        raise PipelineConfigError(
                            "The provided config is not valid JSON"
                        )

            elif args.config is not None:
                try:
                    config = json.loads(args.config)
                except json.JSONDecodeError:
                    raise PipelineConfigError("The provided config is not valid JSON")
            else:
                config = {}

        self.run(config)



from openhexa.sdk import current_run, pipeline


@pipeline("skeleton-pipeline-code", name="Skeleton pipeline name")
def skeleton_pipeline_name():
    count = task_1()
    task_2(count)


@skeleton_pipeline_name.task
def task_1():
    current_run.log_info("In task 1...")

    return 42


@skeleton_pipeline_name.task
def task_2(count):
    current_run.log_info(f"In task 2... count is {count}")


if __name__ == "__main__":
    skeleton_pipeline_name()
